@@ -1,7 +1,7 @@
 #
 # SessionRMarkdown.R
 #
-# Copyright (C) 2020 by RStudio, PBC
+# Copyright (C) 2021 by RStudio, PBC
 #
 # Unless you have received this program directly from RStudio pursuant
 # to the terms of a commercial license agreement with RStudio, then
@@ -155,9 +155,26 @@
       list()
     }
   )
+  
+  haveQuarto <- function() {
+    nzchar(Sys.which("quarto"))
+  }
+
+  isQuartoDoc <- function() {
+     # .qmd file
+     .rs.endsWith(file, ".qmd") ||
+     # file with "format" yaml and no "output" yaml
+     (is.null(yamlFrontMatter[["output"]]) && !is.null(yamlFrontMatter[["format"]])) ||
+     # quarto extended type
+     identical(.Call("rs_detectExtendedType", file), "quarto-document") ||
+     # plain markdown file w/ "jupyter" metdata
+     (.rs.endsWith(file, ".md") && !is.null(yamlFrontMatter[["jupyter"]]))
+  }
 
   if (is.character(yamlFrontMatter[["knit"]]))
     yamlFrontMatter[["knit"]][[1]]
+  else if (haveQuarto() && isQuartoDoc())
+     "quarto render"
   else if (!is.null(yamlFrontMatter$runtime) &&
            grepl('^shiny', yamlFrontMatter$runtime)) {
     # use run as a wrapper for render when the doc requires the Shiny runtime,
@@ -224,8 +241,9 @@
    current <- fileExists && 
       file.info(outputPath)$mtime >= file.info(target)$mtime
    
+   # return named list and alias paths (this data goes to the client)
    list(
-      output_file = .rs.scalar(outputPath),
+      output_file = .rs.scalar(.rs.createAliasedPath(outputPath)),
       is_current  = .rs.scalar(current),
       output_file_exists = .rs.scalar(fileExists)
    )
@@ -237,7 +255,8 @@
 
 # given a path to a folder on disk, return information about the R Markdown
 # template in that folder.
-.rs.addFunction("getTemplateYamlFile", function(path) {
+.rs.addFunction("getTemplateYamlFile", function(path)
+{
    # check for required files
    templateYaml <- file.path(path, "template.yaml")
    skeletonPath <- file.path(path, "skeleton")
@@ -247,11 +266,13 @@
          return(NULL)
    }
 
-   if (!file.exists(file.path(skeletonPath, "skeleton.Rmd")))
+   # validate that a skeleton.Rmd file exists
+   paths <- file.path(skeletonPath, c("skeleton.rmd", "skeleton.Rmd"))
+   if (!any(file.exists(paths)))
       return(NULL)
 
    # will need to enforce create_dir if there are multiple files in /skeleton/
-   multiFile = length(list.files(skeletonPath)) > 1 
+   multiFile <- length(list.files(skeletonPath)) > 1
 
    # return metadata; we won't parse until the client requests template files
    list(
@@ -260,7 +281,6 @@
    )
 })
 
-
 .rs.addFunction("evaluateRmdParams", function(contents) {
 
    Encoding(contents) <- "UTF-8"
@@ -445,17 +465,16 @@
   return(.rs.getRmdOutputInfo(target))
 })
 
-.rs.addFunction("inputDirToIndexFile", function(input_dir) {
-   index <- file.path(input_dir, "index.Rmd")
-   if (file.exists(index))
-      index
-   else {
-      index <- file.path(input_dir, "index.md")
-      if (file.exists(index))
-         index
-      else
-         NULL
-   }
+.rs.addFunction("inputDirToIndexFile", function(input_dir)
+{
+   paths <- c(
+      file.path(input_dir, "index.Rmd"),
+      file.path(input_dir, "index.md")
+   )
+   
+   for (path in paths)
+      if (file.exists(path))
+         return(path)
 })
 
 .rs.addFunction("getAllOutputFormats", function(input_dir, encoding) {
@@ -523,6 +542,16 @@
 
 .rs.addFunction("tinytexRoot", function()
 {
+   # check for tlmgr path set via option; the tinytex root directory is
+   # always 2 directories up from that
+   tlmgr <- getOption("tinytex.tlmgr.path", default = NULL)
+   if (!is.null(tlmgr))
+   {
+      root <- dirname(dirname(dirname(tlmgr)))
+      return(root)
+   }
+   
+   # otherwise, use default locations for different platforms
    sysname <- Sys.info()[["sysname"]]
    if (sysname == "Windows")
       file.path(Sys.getenv("APPDATA"), "TinyTeX")
@@ -534,11 +563,7 @@
 
 .rs.addFunction("tinytexBin", function()
 {
-   root <- tryCatch(
-      tinytex:::tinytex_root(),
-      error = function(e) .rs.tinytexRoot()
-   )
-   
+   root <- .rs.tinytexRoot()
    if (!file.exists(root))
       return(NULL)
    
@@ -564,10 +589,6 @@
 
 .rs.addFunction("bookdown.renderedOutputPath", function(websiteDir, outputPath)
 {
-   # set encoding
-   Encoding(websiteDir) <- "UTF-8"
-   Encoding(outputPath) <- "UTF-8"
-   
    # if we have a PDF for this file, use it
    if (tools::file_ext(outputPath) == "pdf")
       return(outputPath)
